function GCBatchManager(contacts)
{
	this.contacts = contacts;
	this.addEntry = new Array();
	this.delEntry = new Array();
	this.modEntry = new Array();
	this.allEntry = new Array();
	this.submitted = false;
	this.contactBatchLink = "";
	this.groupBatchLink = "";
	this.doneGroup = false;
	this.doneContact = false;
	//this.batchMax = 100;
	this.batchMax = 90;
	//this.batchMaxByte = 1048576; //1MB
	this.batchMaxByte = 1000000; //1MB - mergin
	
	this.atom = Namespace("", "http://www.w3.org/2005/Atom");
	this.openSearch = Namespace("openSearch", "http://a9.com/-/spec/opensearch/1.1/");
	this.gContact = Namespace("gContact", "http://schemas.google.com/contact/2008");
	this.gd = Namespace("gd", "http://schemas.google.com/g/2005");
	this.batch =  Namespace("batch", "http://schemas.google.com/gdata/batch");
	this.localeBundle = document.getElementById("GmContBundle");
}

GCBatchManager.prototype.init = function() {
	this.addEntry = new Array();
	this.delEntry = new Array();
	this.modEntry = new Array();
	this.allEntry = new Array();
	this.submitted = false;
	//this.contactBatchLink = "";
	//this.groupBatchLink = "";
	this.doneGroup = false;
	this.doneContact = false;
}

GCBatchManager.prototype.setContactBatchLink = function(link)
{
	this.contactBatchLink = link;
}

GCBatchManager.prototype.setGroupBatchLink = function(link)
{
	this.groupBatchLink = link;	
}

GCBatchManager.prototype.registerAddEntry = function(entryObj)
{
	this.addEntry.push(entryObj);
	this.allEntry.push(entryObj);
}

GCBatchManager.prototype.registerDelEntry = function(entryObj)
{
	this.delEntry.push(entryObj);
	this.allEntry.push(entryObj);
}

GCBatchManager.prototype.registerModEntry = function(entryObj)
{
	this.modEntry.push(entryObj);
	this.allEntry.push(entryObj);
}

GCBatchManager.prototype.confirmDeletion = function() 
{
	if (this.delEntry.length == 0) return;
	
	var items = new Array();
	for (var i=0; i<this.delEntry.length; i++) {
		var contact = this.delEntry[i];
		var label = "";
		if (contact instanceof GCContactEntry) {
			var addr = contact.getPrimaryAddress();
			var name = contact.getDispName();
			label = name + " <" + addr + ">";
		} else {
			label = contact.getDispName();
		}
		var item = {
			label: label,
			index: i,
			checked: true
		};
		items.push(item);
	}
	//var msg = "The following contacts/groups in %1 will be removed. Check contacts you want to delete.";
	var msg = this.localeBundle.getString("gmcont.conf_batchdel_gc");
	msg = msg.replace("%S", this.contacts.account);
	var title = this.localeBundle.getString("gmcont.conf_batchdel_title_gc");
	var callback = {
		title: title,
		items: items,
		desc: msg,
		isOK: false,
		rows: 10
	};
	
	var windowMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].
												getService(Components.interfaces.nsIWindowMediator);
	var addressbook = windowMediator.getMostRecentWindow("mail:addressbook");
	var win = addressbook ? addressbook : window;
	win.openDialog("chrome://gmcont/content/confirmContactsForm.xul", "gmcont-alert", "chrome,modal,dialog,centerscreen", callback);
	nsPreferences.setBoolPref("gmcont.confirm_batchdel", !callback.notShowAgain);
	
	var newDelEntry = new Array();
	if (callback.isOK) {
		for (var i=0; i<items.length; i++) {
			var item = items[i];
			if (item.checked) newDelEntry.push(this.delEntry[item.index]);
		}
	}
	this.delEntry = newDelEntry;
}

GCBatchManager.prototype.doBatch = function()
{
	if (this.addEntry.length == 0 && this.delEntry.length == 0 && this.modEntry.length ==0) {
		return false;
	}
			
	if (this.addEntry.length == 0 && this.delEntry.length == 0) {
		return this.doBatchUpdate();
	}
	
	dump("Batch add and delete\n");
	//add and delete first
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	var batch = this.batch;
	
	var contactFeed = this.createBatchFeed("contact");
	var groupFeed = this.createBatchFeed("group");
	
	var obj = null;
	var i = 0;
	var needPostContact = false; 
	var needPostGroup = false;
	while (i<this.batchMax && (obj = this.addEntry.shift())) {
		var entry = obj.entry;
		entry.batch::id = "add"+i;
		entry.batch::operation = "";
		entry.batch::operation.@type = "insert";
		if (obj instanceof GCContactEntry) {
			if (!this.appendEntry(contactFeed, entry)) {
				this.addEntry.unshift(obj);
				break;
			}
			//contactFeed.atom::entry += entry;
			needPostContact = true;
			i++;
		}	else if (obj instanceof GCGroupEntry) {
			if (!this.appendEntry(groupFeed, entry)) {
				this.addEntry.unshift(obj);
				break;
			}
			//groupFeed.atom::entry += entry;
			needPostGroup = true;
			i++;
		}		
	}
	
	obj = null;
	//i = 0;
	while (i<this.batchMax && (obj = this.delEntry.shift())) {
		var id = obj.getId();
		var delLink = obj.getEditLink();
		var entry = <atom:entry xmlns:atom={atom} xmlns:gd={gd} xmlns:openSearch={openSearch} xmlns:gContact={gContact} xmlns:batch={batch}></atom:entry>;
		entry.atom::id = delLink.toString();
		entry.batch::id = id.toString();
		entry.batch::operation = "";
		entry.batch::operation.@type = "delete";
		if (obj instanceof GCContactEntry) {
			if (!this.appendEntry(contactFeed, entry)) {
				this.delEntry.unshift(obj);
				break;
			}
			//contactFeed.atom::entry += entry;
			needPostContact = true;
			i++;
		}	else if (obj instanceof GCGroupEntry && !obj.isSystemGroup()) {
			if (!this.appendEntry(groupFeed, entry)) {
				this.delEntry.unshift(obj);
				break;
			}
			//groupFeed.atom::entry += entry;
			needPostGroup = true;
			i++;
		}
	}
	this.doneGroup = !needPostGroup;
	this.doneContact = !needPostContact;
	if (needPostContact) this.postBatch(contactFeed, false);
	if (needPostGroup) this.postBatch(groupFeed, true);	
	
	//return true;
	return needPostContact || needPostGroup;
}

GCBatchManager.prototype.doBatchUpdate = function()
{
	if (this.modEntry.length == 0) {
		return false;
	}
	dump("Batch update\n");
	
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	var batch = this.batch;
	
	var contactFeed = this.createBatchFeed("contact");
	var groupFeed = this.createBatchFeed("group");
	
	var obj = null;
	var i = 0;
	var needPostContact = false; 
	var needPostGroup = false;
	var modContacts = new Array();
	while (i<this.batchMax && (obj = this.modEntry.shift())) {
		if (obj instanceof GCContactEntry) {
			modContacts.push(obj);
			continue;
		}	else if (obj instanceof GCGroupEntry) {
			var updated = this.contacts.updateGroupForContact(obj, true);
			if (updated.length > 0) modContacts = modContacts.concat(updated);
			if (!obj.isSystemGroup()) {
				var entry = obj.entry;
				entry.batch::operation = "";
				entry.batch::operation.@type = "update";
				//groupFeed.atom::entry += entry;
				if(!this.appendEntry(groupFeed, entry)) {
					this.modEntry.unshift(obj);
					break;
				}
				needPostGroup = true;
				i++;
			}
		}
	}

	//register contacts which belong/unbelong to groups to modEntry
	for (var i=0; i<modContacts.length; i++) {
		var contact = modContacts[i];
		var contact2 = this.findEntry(
												contact.getPrimaryAddress(),
												contact.getDispName(),
												this.modEntry,
												true
									 );
		if (!contact2) {
			this.modEntry.push(contact);
		} else {
			dump("contact is found\n");
		}
	}
	
	i = 0;
	while (i<this.batchMax && (obj = this.modEntry.shift())) {
		if (obj instanceof GCContactEntry) {
			var entry = obj.entry;
			entry.batch::operation = "";
			entry.batch::operation.@type = "update";
			if (!this.appendEntry(contactFeed, entry)) {
				this.modEntry.unshift(obj);
				break;
			}
			//contactFeed.atom::entry += entry;
			needPostContact = true; 
			i++;
		}
	}

	this.doneGroup = !needPostGroup;
	this.doneContact = !needPostContact;
	if (needPostContact) this.postBatch(contactFeed, false);
	if (needPostGroup) this.postBatch(groupFeed, true);
	
	return needPostContact || needPostGroup;
}

GCBatchManager.prototype.appendEntry = function(entries, entry)
{
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	var batch = this.batch;
	entries.atom::entry += entry;
	
	//calculate data size
	var data = entries.toString();
  var size = 0;
  for (var i=0; i<data.length; i++){
    var n = escape(data.charAt(i));
    if (n.length < 4) size++; else size+=2;
  }
  
  var ret = true;
  if (size > this.batchMaxByte) {
  	var i = entries.atom::entry.length() - 1;
  	delete entries[i];
  	ret = false;
  	dump("Info: exceed max bytes\n");
  	alert("Info: exceed max bytes");
  }
  
  return ret;
}

GCBatchManager.prototype.postBatch = function(feed, group)
{
  var httpRequest = new XMLHttpRequest();
  if (group) httpRequest.open("POST", this.groupBatchLink, true);
  else httpRequest.open("POST", this.contactBatchLink, true);
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.contacts.gDataAuthToken);
	//httpRequest.setRequestHeader("GData-Version", "2");
	httpRequest.setRequestHeader("GData-Version", "3.0");
	
	var me = this;
  var recvFunc = function() {
   	dump("[Batch] "+httpRequest.status+": "+httpRequest.statusText+"\n");
		me.contacts.outputMessageToStatusBar("[Batch] " + httpRequest.status+": "+httpRequest.statusText);
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
  			me.parseFeed(httpRequest.responseText, group);
       	break;
      case 401: /* Authorization required. */
      case 403: /* Unsupported standard parameter, or authentication or
                   Authorization failed. */
      case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
      	break;
      default:
      	break;
    	}
   	};
   	
 	//var postBody = feed;
 	var postBody = feed.toXMLString();
  var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
	                  createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
 	converter.charset = "UTF-8";
 	var stream = converter.convertToInputStream(postBody);
 	httpRequest.setRequestHeader("Content-Type", "application/atom+xml");
 	httpRequest.setRequestHeader("If-Match", "*");
	me.contacts.sendHttpRequest(httpRequest, stream, recvFunc);
 	return;
}

GCBatchManager.prototype.createBatchFeed = function(category)
{
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	var batch = this.batch;
	
	var feed = 
		<atom:feed xmlns:atom={atom} xmlns:gd={gd} xmlns:openSearch={openSearch} xmlns:gContact={gContact} xmlns:batch={batch}/>;

  feed.atom::category.@scheme = "http://schemas.google.com/g/2005#kind";
  feed.atom::category.@term = "http://schemas.google.com/contact/2008#"+category;
  
  return feed;
}

GCBatchManager.prototype.findEntry = function(key1, key2, entries, contact)
{
		var mod_key1 = key1.replace(/:\d+$/,""); // for group search
		for (var i=0; i<entries.length; i++) {
			var entry = entries[i];
			if (contact && entry instanceof GCContactEntry) {
				if (entry.getPrimaryAddress() == key1 || entry.getDispName() == key2) {
					return entry;
				}
			} else if (!contact && entry instanceof GCGroupEntry) {
				var maillist = entry.getMailList();
				var name;
				if (maillist) name = maillist.dirName;
				else name = entry.getDispName();
				//var name = entry.getMailList() ? entry.getMailList().dirName : entry.getDispName();
				//remove uniquifier
				name = name.replace(/:\d+$/,"");
				if (name == mod_key1) return entry;
			}
		}
		
		return null;
}

GCBatchManager.prototype.parseFeed = function(feed, isGroup)
{
	// setup namespaces
	var atom = this.atom;
	var openSearch = this.openSearch;
	var gContact = this.gContact;
	var gd = this.gd;
	var batch = this.batch;
	/*
	 * The following statement does not work due to the following bug:
	 * Bug 330572 E4X: using default xml namespace assignment inside of function has no effect
	 * https://bugzilla.mozilla.org/show_bug.cgi?id=330572
	 */
	//default xml namespace = atom;
	
	// remove <?xml ?>
	var contactsXML = new XML(feed.replace(/\<\?xml.+\?\>/,""));
	
	var entries = contactsXML.atom::entry;
	if (isGroup) {
		for (var i = 0; i < entries.length(); i++) {
			var entry = entries[i];
			dump(entry.batch::status.@code+": "+entry.batch::status.@reason+"\n");
			if (entry.batch::status.@code != 200 && entry.batch::status.@code != 201) continue;
			var opType = entry.batch::operation.@type.toString();
			delete entry.batch::status;
			delete entry.batch::operation;
			dump("group: "+opType+"\n");
			switch (opType) {
				case "insert":
					delete entry.batch::id;
					var key1 = entry.atom::title.toString();
					var group = this.findEntry(key1, null, this.allEntry, false);
					if (!group) {
						dump("no group was found\n");
						group = new GCGroupEntry(entry);
						this.allEntry.push(group);
					}
					group.init(entry);
					group.setMailList(group.getMailList(), this.contacts.account); //register groupId to pref
					this.contacts.groups[group.getId()] = group;
					this.modEntry.push(group);
					break;
				case "delete":
					//var id = entry.atom::id;
					var batchId = entry.batch::id.toString();
					//delete this.contacts.groups[batchId];
					this.contacts.deleteGroupEntry(this.contacts.groups[batchId]);
					delete entry.batch::id;
					break;
				case "update":
					var id = entry.atom::id.toString();
					var group = this.contacts.groups[id];
					group.init(entry);
					break;
				default:
					break;
			}
		}
		this.doneGroup = true;
	} else {
		for (var i = 0; i < entries.length(); i++) {
			var entry = entries[i];
			dump(entry.batch::status.@code+": "+entry.batch::status.@reason+"\n");
			if (entry.batch::status.@code != 200 && entry.batch::status.@code != 201) continue;
			var opType = entry.batch::operation.@type.toString();
			delete entry.batch::status;
			delete entry.batch::operation;
			dump("contact: "+opType+"\n");
			switch (opType) {
				case "insert":
					delete entry.batch::id;
					var addr = "";
					try {
						//addr = entry.gd::email.(@primary=="true").@address;
						addr = entry.gd::email.(function::attribute("primary")=="true").@address;
					} catch(e) {
						addr = entry.gd::email.@address[0];
					}
					//var dispName = entry.atom::title;
					var dispName = "";
					try {
						dispName = entry.gd::name.gd::fullName.toString();
					} catch(e) {
					}
					var contact = this.findEntry(addr, dispName, this.allEntry, true);
					if (contact) {
						contact.init(entry);
						var abCard = contact.getABCard();
						if (abCard) {
							/*
 							var abURI = this.contacts.getAddressBookURI();
							abCard.setStringAttribute("GmBookUri", abURI);
							abCard.setStringAttribute("GmContactId", entry.atom::id);
							abCard.editCardToDatabase(abURI);
							*/
							abCard.setProperty("GmBookUri", this.contacts.getAddressBookURI());
							abCard.setProperty("GmContactId", entry.atom::id.toString());
							this.contacts.getAddressBook().modifyCard(abCard);
							contact.setABCard(abCard);
						} else {
							contact = new GCContactEntry(entry);
						}
					} else {
						contact = new GCContactEntry(entry);
					}
					this.contacts.contacts[contact.getId()] = contact;
					break;
				case "delete":
					//var id = entry.atom::id;
					var batchId = entry.batch::id.toString();
					//delete this.contacts.contacts[batchId];
					this.contacts.deleteContactEntry(this.contacts.contacts[batchId]);
					delete entry.batch::id;
					break;
				case "update":
					var id = entry.atom::id.toString();
					var contact = this.contacts.contacts[id];
					if (contact.map_del_flag) delete this.contacts.contacts[id];
					else contact.init(entry);
					break;
				default:
					break;
			}
		}
		this.doneContact = true;
 }
 
 if (this.doneContact && this.doneGroup && !this.doBatch()) {
 		this.submitted = true;
 		this.contacts.refreshABCards();
 }
}